//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define FIBER_GUIDE_CGI_C
#include "fiber_guide_cgi.h"

#ifdef __TEST__
FILE *logfile;
#endif // __TEST__
pthread_t thread; //    


mqd_t mq_out = -1;
ssize_t mq_size = 0;
char *out_buf = NULL;
struct mq_attr attr;

void mk_queues(){
	attr.mq_maxmsg = 256; //  256   
	attr.mq_flags = O_NONBLOCK;
	attr.mq_msgsize = 512; //    
	mq_out = mq_open(OUT_QUEUE, O_WRONLY|O_CREAT|O_NONBLOCK, 00666, &attr);
	if(mq_out < 0){printf("Error opening out queue: %s\n", strerror(errno)); exit(1);}
	if(mq_getattr(mq_out, &attr) < 0){
		printf("Error getting queue attributes: %s\n", strerror(errno));exit(1);}
	mq_size = attr.mq_msgsize;
	if((out_buf= (char*)calloc(mq_size, 1)) == NULL){
		printf("Can't allocate memory: %s\n", strerror(errno));exit(1);}
	mq_size--; //     
}

void rm_queues(){
	int i, j=-1;
	MSG("RM_QUEUES\n");
	if(mq_out > 0){
		sleep(2);
		mq_close(mq_out);
	}
	else MSG("No QUEUE\n");
	for(i=0; i<100; i++)
		if((j=mq_unlink(OUT_QUEUE))==0) break;
		else usleep(1000);
	if(j) ERR("Can't remove queue, %s\n", strerror(errno));	
}

void stop(int sig){
	char shmname[32];
	int i;
#ifdef __TEST__
	if(!logfile) fprintf(stderr, "NO LOG!\n");
	else{
		fprintf(logfile, "Caught SIG: %d\n", sig);
		if(sig == SIGUSR1){
			fprintf(logfile, "Attempt to start another process\n");
		}
		fflush(logfile);
	}
#endif // __TEST__
	fprintf(stderr, "Caught signal: %d\n", sig);
	if(sig == SIGINT || sig == SIGQUIT || sig == SIGABRT){
		WARN("  .. ");
		shmctl(shmid, IPC_RMID, NULL); //   *Nbuf_cur
		WARN("OK\n");
#ifndef __NO_MOTORS__
		WARN("  .. ");
		pthread_kill(thread, 9);
		stop_all_devices();
		WARN("OK.\n.\n");
		fclose(outfile);
#endif
		if(out_buf && sig != SIGABRT) free(out_buf);
		for(i=0; i<4; i++){
			snprintf(shmname, 31, "/shm%d", i);
			shm_unlink(shmname);
		}
		rm_queues();
		exit(0);
	}
	else signal(sig, stop);
}

/*
 *	     (shared memory)
 *	    cgi  
 *	   cgi   novideo
 */

struct Params *get_pars_area() {
	char *fname = SHMEM_FILE; // ,     
	int fsize = sizeof(struct Params); //   
	int i, fd=-1, new=0; 
	char *buf, msg[100];
	if((fd = open(fname, O_RDWR)) < 0){ //    
		new = 1;
	}
	else{
		struct stat st;
		fstat(fd, &st);
		if(st.st_size != fsize){
			MSG("Struct Params maybe changed.. (filesize: %ld, st_size: %d)\n", st.st_size, fsize);
			close(fd);
			unlink(fname);
			new = 1;
		}
	}
	if(new){
		if((fd = open(fname, O_CREAT|O_RDWR, 0666)) < 0) { // 
			sprintf(msg,"Can't create file:%s", fname);
			perror(msg);
		}
		else{
			chmod(fname, 0666);
			char bb = 0;
			for(i=0; i<fsize; i++)
				write(fd, &bb, 1); //   ..
			new=1; //..  
		}
	}
	if((buf = mmap(NULL, fsize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0)) == MAP_FAILED ){
		sprintf(msg,"Can't map file:%s(%d) to memory",fname,fsize);
		perror(msg); //   
		fd=-1;
	}
	if(fd<0){ //    ,   
		buf=malloc(sizeof(struct Params));
		new=1;
	}
	pars=(struct Params *)buf;
	if(new){ //     ,    
		MOVETOCENTER = 0;
		Guiding = 0;
		HistFilt = 0;
		FrameS = 64;
		Xtarg  = xtarg;
		Ytarg  = ytarg;
		Xprev = CH_OBJ;
		Yprev = CH_OBJ;
		Nadd = 1;
		FrameRate = 1/(dt_live+0.1);
		RealFrameRate = 0;
		GrabChan  = GRAB_SOURCE;
		Grab_Width = MAX_WIDTH;
		Grab_Height= MAX_HEIGHT;
		XSteps = 0;
		YSteps = 0;
		Brightness = 50.0;
		Contrast = 30.0;
		F_TYPE = No;
		Reset = 0;
		Init = 0;
		ComPort = 0;
		Lamp = 0;
		Shutter = 0;
		Ch_obj = 0;
		CMass = 0;
		dX = 0; dY = 0;
		dCam = 0; newCamPos = 0; //    
		dGr = 0; newGrPos = 0; //     
		dPh = 0; newPh = 0; //       
		W_filter.treshold = 32;
		W_filter.order = 4;
		F_control = 0;
		ShowHist = 0;
		XHist = 0; histVmin = 0; histVmax = 0;
		Shutter_lock = 0;
	}
	return (struct Params *)buf;
}

int main(int argc, char *argv[]){
	int _, num, realnum=0, nadd=1;
	int nchan=GRAB_SOURCE;
#ifndef NOAUTH
	int AccessLevel = 0;
#endif
	int nport;
	int x0, y0, oldfw=0;
	int mjpg_stream=0, main_proc=0;
	char  *qs = NULL,*rm,*par,*nam,*val, text[256];
	double t; //, dx, dy;
	char *grab_dev=GRAB_DEVICE;
	setbuf(stdout, NULL); //    
#ifndef NOAUTH
	if((AccessLevel = get_auth_level()) < SpectatorLevel) exit(-1);
#endif
	mk_queues();
	pars = get_pars_area(); //      
	if((rm = getenv("REQUEST_METHOD")) != NULL &&       // CGI method POST
		strcmp(rm,"POST") == 0)
			qs = fgets(text, 255, stdin); //   POST
	else if((qs=getenv("QUERY_STRING"))!=NULL)  // CGI method GET
		qs = strncpy(text,qs,255);
	else if(argc > 1) {                         // daemon (no CGI)
			qs = strncpy(text, argv[1], 255);
			mjpg_stream=0;
		}
	if(qs == NULL){
		printf("Content-type: text/html; charset=koi8-r\n\n");
		sendJavaScript(NULL, 1);
		exit(0);
	}
#ifndef NOAUTH
	if(AccessLevel < FullLevel)
#endif
#ifndef NOVIDEO
		if(strstr(qs, "video=yes")){ //     ,  
			MSG("Start streaming\n");
			start_streaming();
			exit(0);
		}
#endif
pthread_mutex_lock(&mutex); //      
	par = strtok(qs,"&,; |\n");
	do { //   
		if(strcmp(par, "init") == 0){ //  
			main_proc = 1;
			continue;
		}
		if(strcmp(par, "short") == 0){ //         
			printf("Content-type: text/html; charset=koi8-r\n\n");
			sendJavaScript(NULL, 0);
			exit(0);
		}
		if(strcmp(par, "reset") == 0){ // () 
			Init = 1; //  Init    
			continue;
		}
		if((val = strchr(par, '=')) != NULL)// continue;
			*val++ = '\0';
		nam = par;
		// video=yes, mjpeg=yes  mjpg=yes  
		if(strcmp(nam,"video")==0||strcmp(nam,"mjpeg")==0||strcmp(nam,"mjpg")==0)
			mjpg_stream = (strcmp(val,"yes")==0);
		// nchan -    
		else if(strcmp(nam,"nchan")==0) GrabChan=abs(atoi(val));
		// rate/framerate - -   ,   
		else if(strcmp(nam,"rate")==0||strcmp(nam,"framerate")==0){
			double framerate = fabs(atof(val));
			if(framerate < 0.5) framerate = 0.5; //   1   2 
			dt_live = 1.0/framerate - 0.1;
			if(dt_live < 0.1) dt_live = 0.1; //   10 /
			FrameRate = 1.0/(dt_live + 0.1);
		}
		//  ,      
		else if(strcmp(nam,"framesize")==0){
		    int tmpi = abs(atoi(val));
		    if(FrameS != tmpi){ FrameS = tmpi; Xprev = CH_OBJ; Ch_obj = 1;}
		}
		//   
		else if(strcmp(nam,"xtarg")==0){
		    int tmpi = abs(atoi(val));
		    if(ShowHist){XHist = tmpi;}
			else if(Xtarg != tmpi){Xtarg = tmpi; Xprev = CH_OBJ; Ch_obj = 1;}
		}
		else if(strcmp(nam,"ytarg")==0){
		    int tmpi = abs(atoi(val));
		    if(Ytarg != tmpi && !ShowHist){ Ytarg = tmpi; Yprev = CH_OBJ; Ch_obj = 1;}
		}
		else if(strcmp(nam,"nadd")==0){
			Nadd = atoi(val);
			if(Nadd < 1 || Nadd > 150) Nadd = 1;
		}
		else if(strcmp(nam,"brightness")==0) grab_adjust_brightness(atof(val));
		else if(strcmp(nam,"contrast")==0) grab_adjust_contrast(atof(val));
		else if(strcmp(nam,"histfilt")==0) HistFilt = atoi(val);
		else if(strcmp(nam,"Guiding")==0) Guiding=atoi(val);
		else if(strcmp(nam,"moveto")==0) MOVETOCENTER = 1;
		else if(strcmp(nam,"killall")==0 && atoi(val)==0){
			printf("Content-type: text/html; charset=koi8-r\n\n");
			killbrothers(1); // ,    
		}
		else if(strcmp(nam,"CMass")==0) CMass = atoi(val);
		else if(strcmp(nam, "comport")==0) ComPort = atoi(val);
		else if(strcmp(nam,"Lamp")==0){
			int l_=atoi(val); if(l_<0 || l_ > 15) l_=0;Lamp = l_;}
		else if(strcmp(nam,"Shutter")==0) Shutter = atof(val);
		else if(strcmp(nam,"dX")==0) dX = atoi(val);
		else if(strcmp(nam,"dY")==0) dY = atoi(val);
		else if(strcmp(nam,"dCam")==0) dCam = atoi(val);
		else if(strcmp(nam,"setCam")==0) newCamPos = atoi(val);
		else if(strcmp(nam,"dGr")==0) dGr = atoi(val);
		else if(strcmp(nam,"setGr")==0) newGrPos = atoi(val);
		else if(strcmp(nam,"dPh")==0) dPh = atoi(val);
		else if(strcmp(nam,"setPh")==0) newPh = atoi(val);	
		else if(strcmp(nam,"filter_type")==0) F_TYPE = atof(val);
		else if(strcmp(nam,"wavelet_tresh")==0) W_filter.treshold = atoi(val);
		else if(strcmp(nam,"wavelet_order")==0) W_filter.order = atoi(val);
		else if(strcmp(nam,"show_filter")==0) F_control = atoi(val);
		else if(strcmp(nam,"show_hist")==0) ShowHist = atoi(val);
	/*else if (strcmp(nam,"xxxx")==0)   CGI-  */
	}while((par = strtok(NULL,"&,; |\n"))!=NULL);
	if(FrameS>Grab_Width){
		MSG("\n\e[1;33mNew FrameS=%d\e[0m\n", Grab_Width);
		FrameS = Grab_Width;
	}
	if(FrameS>Grab_Height){
		MSG("\n\e[1;33mNew FrameS=%d\e[0m\n", Grab_Height);
		FrameS = Grab_Height;
	}
	if(FrameS<32) FrameS = 32;
	if(FrameS>256) FrameS = 256;
	if(Xtarg<0) Xtarg = 0;
	else
		if(Xtarg>Grab_Width-1) Xtarg = Grab_Width-1;
	if(Ytarg<0) Ytarg = 0;
	else
		if(Ytarg>Grab_Height-1) Ytarg = Grab_Height-1;
pthread_mutex_unlock(&mutex); //  
#ifndef NOVIDEO
	if(mjpg_stream){ //     
		MSG("Start streaming\n");
		start_streaming();//   
		exit(0); //      - 
	}
#endif
	printf("Content-type: text/html; charset=koi8-r\n\n");
#ifndef NOAUTH
	if(AccessLevel < FullLevel) exit(0);
#endif
	if(!main_proc) exit(0); //   
	MSG("main_proc=%d, mjpg_stream=%d\n", main_proc, mjpg_stream);
	killbrothers(0); //   
	sendJavaScript("init\nkillall=1", 1); //    ,   
	for( _=0; _<255; _++)
		signal(_, stop); //   ,  QUIT  TERM 
#ifndef NODAEMON
	close(0); close(1); close(2);
	if(fork() != 0) exit(0);	
#endif
	killbrothers(2); //  PID-
	#ifndef __NO_MOTORS__ //    	
	pthread_create(&thread, NULL, controllers_thread, NULL);
	#endif
#ifdef __TEST__
	logfile = fopen("positions", "w");
	chmod("positions", 0666);
	setbuf(logfile, NULL);
	fprintf(stderr, "log\n");
#endif //__TEST__
pthread_mutex_lock(&mutex);
	MOVETOCENTER = 0;
	Guiding = 0;
	Xprev = CH_OBJ; Yprev = CH_OBJ; //     
	Ch_obj = 1;
pthread_mutex_unlock(&mutex);
	HistFilt = 0;
	brightness = Brightness;
	contrast = Contrast;
#ifndef NOVIDEO
	grab_init(grab_dev);
#endif
	time0 = dtime(); //   
	width = Grab_Width;
	height = Grab_Height;
	for (num=0,realnum=0;;num++) { //  
		int n;
#ifndef NOVIDEO
		if(nchan!=GrabChan) {
			nchan=GrabChan;
			grab_set_chan(nchan);
		}
		if(nadd != Nadd){
			nadd = Nadd;
			realnum = 0;
			time0 = dtime();
		}
		if(contrast != Contrast){
			change_controls(0, (int)(Contrast/100.*65535));
			contrast = Contrast;
		}
		if(brightness != Brightness){
			change_controls((int)(Brightness/100.*65535), 0);
			brightness = Brightness;
		}
/*		if(realnum > 300){
			realnum = 0; //  ~300     
			time0 = dtime(); //   
		}*/
		//       
		fw = (FrameS<width-1)?FrameS:width-2;
		//     +  -  
		if(fw > height) fw = height;
		int pow2 = 0;
		while(fw != 1){
			fw >>= 1;
			pow2++;
		}
		fw = fh = (1 << pow2);
		 //      / :
		if(fw != oldfw || num == 0){
			Realloc_memory(fw);
			oldfw = fw;
			realnum = 0;
			time0 = dtime();
		}
		x0 = Xtarg-fw/2;
		y0 = Ytarg-fh/2;
		if(x0<0) x0 = 0;
		else if(x0+fw>=width) x0 = width-fw-1;
		if(y0<0) y0 = 0;
		else if(y0+fh>=height) y0 = height-fh-1;	
		MSG("grabbing frame.. ");
		image = grab_next(&width, &height);
		t = dtime();
		if(image==NULL){
			MSG("NO IMAGE!\n");
			continue;
		}
		MSG("done, format=%dx%d\n", width, height);
		if(Ch_obj){
			realnum = 0; num = -1;
			MSG("\n\e[2;30;41mFrame size (maybe) changed, FrameS=%d, \
				\e[1;31;40m", FrameS);
			MSG("\ttime: %f\e[0m\n", ltime());
			Ch_obj = 0;
			continue;
		}
		if(num == 0) continue;
		if(Nadd == 1) copy_image(image, width, x0, y0, fw);
		else{ if((num % Nadd) == 1) copy_image(image, width, x0, y0, fw);
			else add_image(image, width, x0, y0, fw);
		}
		if((num % Nadd) != 0) continue;
		if(realnum == 0){
			MSG("Refrence image\n");
			makeRefImage(F_TYPE, fw);
			time0 = dtime();
			realnum++;
			continue;
		}
		findcenter(F_TYPE, &xcent, &ycent, fw); //     
#endif
#ifdef __TEST__
		fprintf(logfile, "%.2f\t%.2f\n", xcent, ycent);
#endif // __TEST__
#ifdef __NO_MOTORS__   //  xprev, yprev   
		Xprev = xcent; //  
		Yprev = ycent;
#endif
		realnum++; //   ,  
		RealFrameRate = (double)(realnum)/ltime(); // -   
		MSG("\n\n\t\t\e[1;31mtime from start: %f\e[0m\n\n", ltime());
	} //for (num=0,realnum=0;;num++)
}

void sendJavaScript(char *msg, bool sflag){
	if(msg) printf("%s\n", msg);
	if(sflag)
	printf(
		"filter_type=ftype%d\n"
		"wavelet_order=w_ord%d\n"
		"wavelet_tresh=w_tres%d\n"
		"nadd=%d\n"
		"xtarg=%d\n"
		"ytarg=%d\n"
		"framesize=fs%d\n"
		"nchan=nch%d\n"
		"framerate=r%d\n"
		"bright_=%.1f\n"
		"contr_=%.1f\n"
		"CMass=CMass%d\n"
		"histfilt=%d\n"
		"comport=com%d\n"
		"Lamp=Lamp%d\n"
		"setCam=%d\n"
		"setGr=%d\n"
		"setPh=%d\n"
		"show_filter=%d\n"
		"show_hist=%d\n"
		,F_TYPE
		,W_filter.order
		,W_filter.treshold
		,Nadd
		,Xtarg
		,Ytarg
		,FrameS
		,GrabChan
		,(int)(FrameRate*10.)
		,Brightness
		,Contrast
		,CMass
		,HistFilt
		,ComPort
		,Lamp
		,CamPos
		,GrPos
		,Ph
		,F_control
		,ShowHist
	);
	printf(
		"realfrm=%.1f\ncentfrm=(%.1f, %.1f)\n"
		"Guiding=%d\nmoveto=%d\n"
		"Shutter=shtr%d\n"
		,RealFrameRate, xcent+Xtarg, ycent+Ytarg
		,Guiding, MOVETOCENTER
		,Shutter
	);
	if(Shutter_lock)
		printf("expos= : %d .\n", ExpTime);
	else printf("expos=\n");
}

double dtime() {
	struct timeval ct;
	struct timezone tz;
	gettimeofday(&ct, &tz);
	return (ct.tv_sec + ct.tv_usec/1e6);
}

double ltime(){ // ,     
	return(dtime()-time0);
}

