//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

/*
 *	    :  ,  ,
 *	 ,   
 */
#define IMG_FUNCTIONS_C
#include "fiber_guide_cgi.h"

/*
		     
*/
int stat_image(unsigned char *image, int x0, int y0, int frame_size){
	int i,j, w=Grab_Width;
	int xm = x0 + frame_size;
	int ym = y0 + frame_size;
	double d, avr, std;
	unsigned char *bp, pix, vmin, vmax;
	unsigned int *hist = NULL;
	unsigned char notv = 0x23;
#ifdef __MORE_TEST__
	double t0 = dtime();
#endif
	MSG("stat_image():");
	vmax = 0;
	for(j=y0; j<ym; j++){
		bp = &image[w * j + x0];
		for(i=x0; i<xm; i++, bp++)
		    if(vmax < *bp) vmax = *bp;
	}
	if(vmax == 0) return 0;
	hist = (unsigned int*) calloc(vmax+1, sizeof(int));
	if(!hist){WARN("     \n"); return 0;}
MSG("\n\nHIST: size=%d, hist[last]=%d\n", vmax, hist[vmax-1]);
	vmin = vmax;
	for(j=y0; j<ym; j++){
		bp = &image[w * j + x0];
		for(i=x0; i<xm; i++, bp++){
			pix = *bp;
			hist[pix]++;
			if(vmin>pix) vmin = pix;			
		}
	}
	if(vmax==vmin && vmax==notv){ /* No TV-signal?!*/
MSG("\n\nFREE HIST (NO TV)\n\n");
		if(hist) free(hist);
		return 0;
	}
	for (avr=0.,i=0,j=vmin; j<=vmax; j++) {
		avr += (double)hist[j]*j; //       
		i += hist[j];		//      
	}
	if(avr > 1. && i != 0) avr /= i; //   
	for(std=0.,j=vmin; j<=vmax; j++){
		d = j-avr;		//   
		std += d*d*hist[j];	//   
	}
	if(i != 0) std = sqrt(std/i);		//  
MSG("\n\nFREE HIST: size=%d, hist[last]=%d\n", vmax, hist[vmax-1]);
	if(hist) free(hist);
	MSG(" time=%f\n", dtime()-t0);
	MSG("Stat: avr=%.2f, std=%.2f, vmin=%d, vmax=%d\n", avr, std, vmin, vmax);
	if(avr < MIN_AVR_LEVEL || std < MIN_STD_LEVEL) return 0;
	else return 1;
}
//      image   img
void copy_image(unsigned char *image,  int width, int x0, int y0, int size){
	int i, j, k=0;
	int xm = x0 + size;
	int ym = y0 + size;
	unsigned char *bp;
	double *im = img;
	MSG("copy_image.. ");
	if(!stat_image(image, x0, y0, size)){
		MSG("No videosignal? Fill image by zeros.\n");
		i = size * size;
		for(j=0; j<i; j++) *im++ = 0.;
		return;
	}
	for(j=y0; j<ym; j++){
		k = width * j + x0;
		for(i=x0; i<xm; i++){
		    *im++ = (double) image[k++];
		}
	}
	MSG("img[0]=%.1f, image[%d]=%d\n", img[0], width*y0+x0, image[width*y0+x0]);
}

void add_image(unsigned char *image,  int width, int x0, int y0, int size){
	int i, j, k=0;
	int xm = x0 + size;
	int ym = y0 + size;
	unsigned char *bp;
	double *im = img;
	if(!stat_image(image, x0, y0, size)){
		MSG("No videosignal? Do not add..\n");
		return;
	}
	MSG("add_image.. ");
	for(j=y0; j<ym; j++){
		k = width * j + x0;
		for(i=x0; i<xm; i++){
		    *im++ += (double) image[k++];
		}
	}
	MSG("img[0]=%d, image[%d]=%d\n", img[0], width*y0+x0, image[width*y0+x0]);
}

//  
void grab_adjust_brightness(double adjval){
	if(adjval>0.1 && Brightness<69.9) {
		Brightness += adjval;
		if(Brightness>70.) Brightness=70.;
	} else if(adjval<-0.1 && Brightness>10.1) {
		Brightness += adjval;
		if(Brightness<10.) Brightness=10.;
	} else if(Brightness>70.) Brightness=70.;
	else if(Brightness<10.) Brightness=10.;
	else return;
	MSG("Adjust brightness to %5.1f%%  \n",Brightness);
}

//  
void grab_adjust_contrast(double adjval){
	if(adjval>0.1 && Contrast<99.9) {
		Contrast += adjval;
		if(Contrast>100.) Contrast=100.;
	} else if(adjval<-0.1 && Contrast>10.1) {
		Contrast += adjval;
		if(Contrast<10.) Contrast=10.;
	} else if(Contrast>100.) Contrast=100.;
	else if(Contrast<10.) Contrast=10.;
	else return;
	MSG("Adjust contrast to %5.1f%%  \n",Contrast);
}


