//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define JPEG_STREAMING_C
#include "fiber_guide_cgi.h"

int *Hptr = NULL; //  ,   

void exit_(int sig){
	if(sig == SIGINT || sig == SIGQUIT || sig == SIGABRT){
		print_last_boundary();
		fflush(NULL);
		exit(0);
	}
	signal(sig, exit);
}
/*
	     mjpeg-
*/
void print_first_header(){
	if(printf("Content-type:  multipart/x-mixed-replace; boundary=myboundary\r\n") < 1) exit(1);
	printf("\r\n--myboundary\r\n");
	printf("Content-type: image/jpeg\r\n");
	if(jpeg_size>0)
		printf("Content-Length: %d\r\n\r\n",jpeg_size);
	else
		printf("\r\n");
}

void print_next_boundary(){
	if(printf("\r\n\r\n--myboundary\r\n") < 1) exit(1);
}

void print_next_header(){
	if(printf("Content-type: image/jpeg\r\n") < 1) exit(1);
	if(jpeg_size>0)
		printf("Content-Length: %d\r\n\r\n",jpeg_size);
	else
		printf("\r\n");
}

void print_last_boundary(){
	printf("\r\n\r\n--myboundary--\r\n");
}

//    
void fill_frame(){
	int i, j, s = FrameS;
	int fs2 = s / 2;
	int x0 = Xtarg - fs2;
	if(x0 < 0) x0 = 0;
	else if(x0 > (Grab_Width - 1)) x0 = Grab_Width - 1;
	int x = x0 + s;
	if(x > (Grab_Width - 1)){
		x = Grab_Width - 1;
		x0 = x - s;
	}
	int y0 = Ytarg - fs2;
	if(y0 < 0) y0 = 0;
	else if(y0 > (Grab_Height - 1)) y0 = Grab_Height - 1;
	int y = y0 + s;
	if(y > (Grab_Height - 1)){
		y = Grab_Height - 1;
		y0 = y - s;
	}
	unsigned int *iptr;
	double *fptr = filterbuf;
	s = Grab_Width;
	for(i = y0; i < y; i++){
		iptr = &ima[i*s+x0];
		for(j = x0; j < x; j++)
			*iptr++ = (unsigned int)1.5 * fabs(*fptr++); //      
	}
}

void plot_histogram(unsigned int *hist, int vmax, int maxh){
	int i, y, oldy = Grab_Height, newXH = XHist;
	double w =  ((double)Grab_Width) / ((double)vmax); //  
	double dh = ((double)Grab_Height) / ((double)maxh); //   
	double x = 0., oldx = 0.;
	double sum = 0.;
	if(newXH != histVmin && newXH != histVmax){
		if(Hptr){ //  ,   
			*Hptr = newXH;
			//    
			if(histVmin > histVmax){int tmp = histVmax; histVmax = histVmin; histVmin = tmp;}
			if((histVmax-histVmin) < 10){ //    
				if(histVmin > 10) histVmin = histVmax - 10;
				else histVmax = histVmin + 10;
			}
			XHist = histVmin;
			Hptr = NULL;
		}
		else{ //  
			int dmin = abs(newXH - histVmin);
			int dmax = abs(newXH - histVmax);
			Hptr = (dmin < dmax) ? &histVmin : &histVmax;
			XHist = histVmin;
		}
	}
	if(w < 2.) dh *= w/2.;
	for(i = 0; i <= vmax; i++){
		if((x - oldx) < 2.){
			sum += (double)(*hist++);
		}
		else{
			y = Grab_Height - (int)(dh * sum);
			gdImageLine(im, (int)oldx, oldy, (int)x, y, red);
			sum = (double)(*hist++);
			oldx = x;
			oldy = y;
		}
		x += w;
	}
	if(Hptr != &histVmin) gdImageLine(im, histVmin, 0, histVmin, Grab_Height, green);
	if(Hptr != &histVmax) gdImageLine(im, histVmax, 0, histVmax, Grab_Height, green);
}


//   , , 
// 		(  )
int prep_frame(){
	int i,j,hd, pix, w=Grab_Width, h=Grab_Height, hbot, htop;
	double d;
	unsigned int *bp, *hist, vmin, vmax;
	unsigned int *ii = ima;
	const int size = w*h;
	const int notv = 0x23;
#ifdef __MORE_TEST__
	double t0 = dtime();
#endif
	int nbot = (int)((double)(size)/20. + 0.5); // 5%   (     )
	int ntop = size - nbot;
	int nmed = size / 2, maxh = 0;
	nbot *= 3;
	MSG("Preparing frame\n");
	if(F_control) fill_frame();
	bp = ima;
	vmax = 0;
	for(i=0; i<size; i++, bp++)
		if(vmax < *bp) vmax = *bp;
	if(vmax == 0){
		return 0;
	}
	hist = (unsigned int*)calloc(vmax+2, sizeof(int));
	bp = ima;
	vmin = vmax;
	MSG("vmax=%d", vmax);
	for(i=0; i<size; i++){
			pix = *bp++;
			hist[pix]++;
			if(vmin>pix) vmin = pix;
		}
	if(vmax==vmin && vmax==notv){ /* No TV-signal?!*/
		free(hist);
		return 0;
	}
	median = 0;
	hbot = 0; htop = vmax;
	for(i=j=0; j<vmax; j++){	//  
		i += hist[j];
		if(maxh < hist[j]) maxh = hist[j];
		if(i<=nmed) median = j; //  
		if(i<=nbot) hbot = j;	//   ~5%   
		else if(i>=ntop){		// -//-  95% -//-
			htop = j;
			break;
		}
	}
	for (avr=0.,i=0,j=vmin; j<=vmax; j++) {
		avr += (double)hist[j]*j; //       
		i += hist[j];		//      
	}
	if(avr > 1. && i != 0) avr /= i; //   
	for (std=0.,j=vmin; j<=vmax; j++) {
		d = j-avr;		//   
		std += d*d*hist[j];	//   
	}
	if(i != 0) std = sqrt(std/i);		//  
	bp = ima;
	if(HistFilt == 1){ //  
		if(histVmin || histVmax){ //  
			double tmp = (double)vmax / Grab_Width;
			hbot = (int)(tmp * histVmin);
			htop = (int)(tmp * histVmax);
		}
		hd = htop - hbot; //   
		if(hd <= 0) hd = 1;
		for(i=0; i<size; i++){
			pix = (int)((double)(*bp-hbot)/(double)hd*255.);
			if(pix>255) pix=255;
			if(pix<0) pix=0;
			*(bp++) = pix;
		}
	}
	else{ //     0..255
		double scale = (double)(vmax) / 254.;
		for(i=0; i<size; i++){ 
			pix = (int)((double)(*bp) / scale);
			*(bp++) = pix;
		}
	}
	for(j=0; j<h; j++)
		for(i=0; i<w; i++)
			im->tpixels[j][i] = levels[*ii++];
	if(ShowHist == 1) plot_histogram(hist, vmax, maxh);
	free(hist);
	MSG("\nprep_frame(): time=%f\n", dtime()-t0);
	return 1;
}

void send_jpeg(){
	int i,j, w=Grab_Width, h=Grab_Height, brect[8], x, y;
//	unsigned int *ii = ima;
	double sz = 22.;
	char *fc = "times", *ss = NULL, *No_sig = "No videosignal?", *Poor_sig = "Poor signal";
	char msg[256];
	MSG("Sending jpeg, size: %dx%d..\n", w, h);
	if(!prep_frame()) ss = No_sig;
	else if(avr < MIN_AVR_LEVEL || std < MIN_STD_LEVEL) ss = Poor_sig;
	if(ss){
		snprintf(msg, 256, "%s std=%.1f, avr=%.1f", ss, std, avr);
		ss = msg;
		MSG("%s\n", ss);
		gdImageStringFT(NULL,brect,red,fc,sz,0.,100.,100.,ss);
		x = (w - brect[2] + brect[0])/2;
		y = (h - brect[1] + brect[7])/2;
		//char *gdImageStringFT(gdImagePtr im, int *brect, int fg, char *fontname, double ptsize, double angle, int x, int y, char *string) 
		gdImageStringFT(im,brect,red,fc,sz,0.,x,y,ss);
	}
	else if(ShowHist==0){
#ifdef __TEST__
		snprintf(msg, 256, "std=%.1f, avr=%.1f; xc=%.1f, yc=%.1f; med=%d",
			std, avr, xcent, ycent, median);
		ss = msg;
		gdImageStringFT(NULL,brect,red,fc,sz,0.,0.,0.,ss);
		x = (w - brect[2] + brect[0])/2;
		y = h - brect[1] + brect[7];
		gdImageStringFT(im,brect,red,fc,sz,0.,x,y,ss);
		snprintf(msg, 256, "stepsX=%d, stepsY=%d; real fr=%.1f, w=%d, h=%d",
			XSteps, YSteps, RealFrameRate, Grab_Width, Grab_Height);
		gdImageStringFT(NULL,brect,red,fc,sz,0.,0.,0.,ss);
		x = (w - brect[2] + brect[0])/2;
		y = brect[1] - brect[7];
		gdImageStringFT(im,brect,red,fc,sz,0.,x,y,ss);
#endif
		//   
		x = Xtarg;
		y = Ytarg;
		i = FrameS/2;
		gdImageSetStyle(im, Dashed, 10);
		gdImageLine(im, x-10, y-10, x+10, y+10, gdStyled);
		gdImageLine(im, x-10, y+10, x+10, y-10, gdStyled);
		//   
		gdImageRectangle(im, x-i, y-i, x+i, y+i, gdStyled);
		//     
		x -= (int)(xcent+0.5); 
		y += (int)(ycent+0.5);
		gdImageSetStyle(im, Dotted, 2);
		gdImageRectangle(im, x-25, y-25, x+25, y+25, gdStyled);
	}
	//  ,    
	if(start){
		print_first_header();
		start=0;
	}
	else
		print_next_header();
	MSG("OK\n");
//	gdImageJpeg(im, stdout, LIVE_JPEG_QUALITY);
	{ //   :
		char *buf; int size;
		buf = (char*)gdImageJpegPtr(im, &size, LIVE_JPEG_QUALITY);
		if(fwrite(buf, size, 1, stdout) <1) exit(1); //   
		gdFree((void *)buf);
	}
	print_next_boundary();
}

#ifdef __SAVE_JPEG__
void writejpg(){
	static FILE *fp = NULL;
	MSG("writejpg.. ");
	mkstemp(tmp_file); //   
	fp = fopen(tmp_file, "w");
	MSG("w=%d, h=%d\n", Grab_Width, Grab_Height);
	gdImageJpeg(im, fp, JPEG_QUALITY);
	fclose(fp);
	MSG("JPEG written\n");
	rename(tmp_file,jpeg_file);
}
#endif // __SAVE_JPEG__

void init_dg_colors(){ //     dg
	int x;
	gdFTUseFontConfig(1);
	im = gdImageCreateTrueColor(Grab_Width, Grab_Height);
	for(x=0; x<256; x++)
		levels[x] = gdImageColorResolve(im, x, x, x);
	red = gdImageColorResolve(im, 255, 0, 0);
	green = gdImageColorResolve(im, 0, 255, 0);
	blue = gdImageColorResolve(im, 0, 0, 255);
	for(x=0; x<5; x++){
		Dashed[x] = red;
		Dashed[x+5] = green;
	}
	Dotted[0] = blue;
	Dotted[1] = red;
}

void j_init_mmap(){
	int i=0, fd;
	key_t tkey;
	char shmname[32];
	struct v4l2_requestbuffers req;
	int filbufsize = 65536 * sizeof(double);
	grab_fd = open (grab_dev, O_RDWR | O_NONBLOCK, 0);
	MSG("Init mmap for jpeg streaming\n");
	if (grab_fd < 0){
		fprintf(stderr, "Cannot open '%s': %d, %s\n",
			grab_dev, errno, strerror (errno));
		exit (EXIT_FAILURE);
	}
	tkey = ftok(grab_dev, 666);
	if( (shmid = shmget(tkey, sizeof(struct Buffs), 0666)) == -1){
		perror("Can't access shared memory");
		exit(0);
	}
	Buffers = (struct Buffs*) shmat(shmid, NULL, 0);
	if(Buffers->nbuf_cur < 0) WARN("      \n");
	while(Buffers->nbuf_cur < 0){// ,      
		i++;  sleep(1);  WARN(" %d ...\n", i);
	}
	for (i = 0; i < 4; ++i){
		snprintf(shmname, 31, "/shm%d", n_buffers);
		if((fd = shm_open(shmname, O_RDWR, 00666))<0){perror("SHM_OPEN");exit(1);};
		images[i] = mmap(NULL, Buffers->length, PROT_READ|PROT_WRITE, 
			MAP_SHARED, fd, 0);
		if(images[i] == MAP_FAILED){
			fprintf (stderr, "Out of memory, len=%d, fd=%d\n", Buffers->length, fd);
			exit (EXIT_FAILURE);
		}
	}
	if((fd = shm_open("/shm5", O_RDWR, 00666))<0){perror("SHM_OPEN");exit(1);};
	filterbuf = mmap(NULL, filbufsize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if(filterbuf == MAP_FAILED) errno_exit("Out of memory");
	MSG("OK, length=%ld\n", Buffers->length);
}

void start_streaming(){
	double framerate_old, frametime = 0.;
	int num, ns, i, j, nbuf_old = -1, *ii, oldw = Grab_Width, oldh = Grab_Height;
	unsigned char *image;
	signal(SIGINT ,exit_);
	signal(SIGQUIT ,exit_);
	signal(SIGABRT ,exit_);
	setbuf(stdout, NULL);
	dt_live = 1.0/FrameRate - 0.1;
	framerate_old = FrameRate;
	init_dg_colors();
	j_init_mmap(); //     
	frametime = dtime();
	j = Grab_Width * Grab_Height;
	for(num=0,ns=0;;){
		//  
		if(oldw != Grab_Width || oldh != Grab_Height){
			gdImageDestroy(im);
			im = gdImageCreateTrueColor(Grab_Width, Grab_Height);
			oldw = Grab_Width; oldh = Grab_Height;
			j = Grab_Width * Grab_Height;
		}
		if(FrameRate != framerate_old){
			dt_live = 1.0/FrameRate - 0.1;
			framerate_old = FrameRate;		
		}
		if(Buffers->nbuf_cur < 0 || Buffers->nbuf_cur == nbuf_old) continue;
		num++;
		nbuf_old = Buffers->nbuf_cur;
		image = images[Buffers->nbuf_cur];
		MSG("nBuf=%d\n", Buffers->nbuf_cur);
		ii = ima;
		if(Nadd == 1 || (num % Nadd) == 0){ //    ,   
			//      
			if((dtime() - frametime) > dt_live){	
				send_jpeg();
				// :  10    
#ifdef __SAVE_JPEG__
				if(((ns++)%10)==0)
					writejpg();
#endif
					frametime = dtime();
				}
			MSG("Prepare image (size=%d)..", j);
			for(i=0; i<j; i++) //   
				*(ii++) = *(image++);
			MSG("OK\n");
		}
		else{
			MSG("Add to previous image..");
			for(i=0; i<j; i++) //   
				*(ii++) += *(image++);
			}
			MSG("OK\n");
	}
}

