//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define MOTOR_FUNC_C
#include "motor-func.h"
#include <stropts.h>
int X_pos = 0, Y_pos = 0; //   ( )
bool Guiding_Lock = 0, oldshtr = 1;
unsigned char oldmask = 0;

double tbeg = 0.; //   

struct termio oldtty, tty; //   UART
int comfd=0;//, slrd; //   

char *cnameByNum(unsigned char cno){
	static char *devnames[] = {"", " ", " Y'", " X'",
		"", "", " ", ""};
	return devnames[cno>>5];
}

unsigned char crc(unsigned char data){
	unsigned char crc = data & 1;
	unsigned int i;
	for(i = 1; i<8; i++) crc ^= (data >> i) & 1;
	return crc;
}

void tty_init(){
	int i, flag = 1;
	unsigned char comdev[32];
	snprintf(comdev, 32, "/dev/ttyS%d", ComPort);
	if(comfd){
		ioctl(comfd,TCSETA,&oldtty);
		close(comfd);
	}
	WARN("  %s... ", comdev);
	if ((comfd = open(comdev,O_RDWR|O_NOCTTY|O_NONBLOCK)) < 0){ //|O_NONBLOCK
		WARN("    %s: %s\n.", comdev, strerror(errno));
		exit(1);
	}
	WARN(" OK\n   ... ");
	if(ioctl(comfd,TCGETA,&oldtty) < 0) flag = 0; //    
	tty = oldtty;
	tty.c_lflag    = 0;
	tty.c_iflag    = BRKINT;
	tty.c_oflag    = 0;
	tty.c_cflag    = B9600|CS8|CREAD|CLOCAL|PARENB;//|CMSPAR;
	tty.c_cflag &= ~PARODD;
	tty.c_cc[VMIN] = 0;
	tty.c_cc[VTIME] = 5;
	if(ioctl(comfd,TCSETA,&tty) < 0) flag = 0;
	if(flag){
		WARN(" OK\n");
	}
	else{
		WARN("   , .\n");
		exit(1);
	}
}

int read_tty_raw(){ //  1  (  )
	unsigned char rb; // ,  
	fd_set rfds; //   
	struct timeval tv; //  
	int retval; //  - select 
	tty.c_iflag &= ~PARODD; //     = 0
	ioctl(comfd,TCSETA,&tty);
	FD_ZERO(&rfds); //  
	FD_SET(comfd, &rfds); //   -  
	tv.tv_sec = 0; tv.tv_usec = 500000; //  .5
	retval = select(comfd + 1, &rfds, NULL, NULL, &tv);
	if (!retval){
		MSG("  !\n");
		return (-1); //   ,  
	}
	if(FD_ISSET(comfd, &rfds)){
		if(read(comfd, &rb, 1) < 1){
			MSG(" \n");
			return (-1); //  
		}
	} 
	else{
		MSG(",   COM-\n");
		return (-1); //  
	}
	MSG(" : %d\n", rb);
	return (int)rb;
}

int read_tty(unsigned char* ctrlr){ //  1  (  )
	int rb=-1, i=0;
	while(rb == -1 && i++ < 3){
		rb = read_tty_raw();
	}
	if(rb < 0) return (-1);
	if(ctrlr != NULL) *ctrlr = rb & 0xE0;
	rb &= 0x1F; //  
	MSG("  %s   %d\n", cnameByNum(*ctrlr), rb);
	return rb;
}

bool write_tty_raw(unsigned char wb){
	if(crc(wb)) //  
		tty.c_cflag |=PARODD; // 9-  = 0
	else
		tty.c_cflag &= ~PARODD; // = 1
	ioctl(comfd,TCSETA,&tty); //  
	MSG(" : %d\n", wb);
	if(write(comfd, &wb, 1) < 0){
		MSG("!    :(\n");
		return 0;
	}else return 1;
}

bool write_tty(unsigned char wb, unsigned char cno){
	unsigned char tmp;
	tmp = wb & 0x1f; //      
	tmp |= cno; //    
	return (write_tty_raw(tmp));
}

bool send_cmd(unsigned char cmd, unsigned char flag, unsigned char cno){ //     
	// flag == 1 -    , 0 -  
	int rtn = -1;
	bool ret = 0;
	unsigned char i = 0, cmd1, ctrlr;
	cmd1 = cmd & 0x1F;
//		usleep(10000);
		if(flag) ret = write_tty(cmd, cno);
		else ret = write_tty_raw(cmd);
		if(!ret){
			WARN("   .\n");
			return 0;
		}
		usleep(10000);
		rtn = read_tty(&ctrlr);
		MSG("rtn=%d (%s)\n", rtn, cnameByNum(ctrlr));
		if(rtn < 0){
			WARN("   .");
			return 0;
		}
//		if(!rtn) rtn = read_tty(&ctrlr);
		if(flag && ctrlr != cno){
//			if(ctrlr == C_Motor) termsig(rtn); //  TERMINAL_
//			else 
			WARN("   :  %d   %s,  %d  %s\n", cmd1, cnameByNum(cno), rtn, cnameByNum(ctrlr));
			return 0;
		}
	if(rtn != cmd1){
		error(rtn, cno);
	}
	else ret = 1;
	MSG("send_cmd: cmd=%d (%s), ans=%d (%s)\n",cmd, cnameByNum(cno), rtn, cnameByNum(ctrlr));
	return ret;
}

bool write_byte(unsigned char byte, unsigned char cno){
	int rr;
	unsigned char ctrlr=8, i=0;
	if(crc(byte)) //  
		tty.c_cflag &= ~PARODD; // 9-  = 0
	else
		tty.c_cflag |= PARODD; // = 1
	ioctl(comfd,TCSETA,&tty); //  
	do{
	 	MSG(" %d... \n",i);
	    if(write(comfd, &byte, 1) < 0) WARN("-    :(\n");
	    usleep(10000);
	    rr = read_tty(&ctrlr);
//	    if(rr && ctrlr != cno){rr = 0; continue;} //  
	    i++;
	}while(rr != OK && i < 2);
	if(rr != OK || ctrlr != cno){
		WARN(" \n");
		return 0;
	}
	return 1;
}

bool sendword(unsigned int data, unsigned char cno){
	unsigned char tmp;
	MSG("  \n");
	tmp = (data >> 8) & 0xFF;
	if(!write_byte(tmp, cno)) return 0;
	MSG("\n  \n");
	tmp = data & 0xFF;
	if(!write_byte(tmp, cno)) return 0;	
	MSG("  %d  .\n\n", data);
	return 1;
}
/*
bool tmrset(unsigned char cno){
	int t_h,t_l;
	float period;
	if(!send_cmd(TMR_SETTINGS, 1, cno))	return 0;
	usleep(1000);
	t_h = read_tty_raw();
	usleep(1000);
	t_l = read_tty_raw();
	if(t_h == -1 || t_l == -1){
		return(0);
	}
	period = (65536.0 - 256.0*(float)t_h - (float)t_l)/125000.0;
	WARN(" : T=%f (%s)\n", period, cnameByNum(cno));
	return 1;
}*/

bool init_ctrlr(unsigned char cno){
	unsigned char ctrlr;
	int rb = -1, i=0;
	WARN("  '%s', ", cnameByNum(cno));
	do{
		WARN(" %d ", i++);
	    if(write_tty(INIT, cno)){ //  
		    rb = read_tty(&ctrlr);
		    if(ctrlr != cno) rb = -1;
		}
	    usleep(10000);
	}
	while(i<10 && rb != INIT); //  
	if(i > 9 || rb != INIT){
		WARN("!!!!!!  :  '%s'   !!!!!!\n", cnameByNum(cno));
		return 0;
	}
	else{
		WARN(" .\n");
		return 1;
	}
}
/*
bool set_timer(unsigned char cno, float time){
	int w2s = (int)(65536. - 125000. * time);
	if(w2s < 0 || w2s > 65533) return 0;
	if(!send_cmd(SET_TIMER, 1, cno)) return 0;
	if(sendword(w2s, cno)){
		WARN("   ");
		return(tmrset(cno));
	}
	return 0;
}*/

int corrector(int steps, const unsigned char dir){
	unsigned char cmd, cno;
	int ret = 0;
	char wrds[20];
	if(steps > 0){
		switch(dir){
			case 0: cmd = STEP_R_NUM; cno = X_Motor; 
				strcpy(wrds, "X ()");
				break; // X+
			case 1: cmd = STEP_L_NUM; cno = X_Motor;
				strcpy(wrds, "X ()");
				break; // X-
			case 2: cmd = STEP_R_NUM; cno = Y_Motor;
				strcpy(wrds, "Y ()");
				break; // Y+ 
			case 3: cmd = STEP_L_NUM; cno = Y_Motor;
				strcpy(wrds, "Y ()");
				break; // Y-
		}
//		send_cmd(SET24V, 1, cno);
		if(send_cmd(cmd, 1, cno))
		if(sendword(steps, cno)){
			 MSG(": %s  %d \n", wrds, steps);
			 ret = 1;
		}
		else WARN(": %s  %d \n", wrds, steps);
//		send_cmd(SET12V, 1, cno);
	}
	return ret;
}

bool read_ctrlr(){
	bool ret = 0;
	unsigned char ctrlr;
	#ifdef __MORE_TEST__
	double t0=dtime();
	#endif
	if(!send_cmd(TERMINALS, 1, C_Motor)) return 0;
	int rb = read_tty(&ctrlr);
	if(rb != -1 && rb != oldmask && ctrlr == C_Motor){
		oldmask = (unsigned char) rb;
		if(rb & 16){ //    
			WARN("   \n");
			Shutter_lock = 0;
			Shutter = 1;
			shutter(1);
		}
		else{
			WARN("   \n");
			Shutter_lock = 1;
			Shutter = 0;
			ExpTime = 0;
			tbeg = dtime();
			shutter(0);
		}	
		usleep(10000);
		ret = 1;}
	MSG("read_ctrlr(), time=%f\n", dtime() - t0);
	return ret;
}

/*
void terminals(unsigned char cno){ //  
	unsigned char ctrlr;
	char term = -1, i=0;
	do{
		term = read_tty(&ctrlr);
		if(ctrlr != cno){term = -1; continue;} //  
	}while(i++ < 3 && ctrlr != cno);
	if(ctrlr != cno) return;
	if(term == -1) return;
	MSG("ctrlr#%d: terminal(s) ", ctrlr);
	if(cno == C_Motor) //    
		termsig(term);
	MSG(" terminals register = %d\n", term);
}*/

void error(unsigned char err, unsigned char cno){ 
	WARN("   '%s': ", cnameByNum(cno));
	switch(err){	
		case STACK_OVERFLOW:
			WARN("  RX.\n"); //   
			break;
		case NO_STOP_BIT:
			WARN("    .\n"); 
			break;
		default:
			WARN("   (%d).\n", err); 
			break;
	}
}
/*
void tty_sig(unsigned char rb, unsigned char cno){ //    
	switch(rb){
		case TERMINAL_: terminals(cno); break; //  
		case STACK_OVERFLOW: error(STACK_OVERFLOW, cno); break; //   RX 
		case NO_STOP_BIT: error(NO_STOP_BIT, cno); break;
	}
//	if(cno == C_Motor) termsig(rb); //  TERMINAL_
}*/

bool gtmid(unsigned char cno){
	bool ret = 0;
//	send_cmd(SET24V, 1, cno);
	if(send_cmd(STEP_L_NUM, 1, cno))
		if(sendword(XYM_steps*2, cno)){ //      
			sleep(1); // ,     
			if(send_cmd(STEP_R_NUM, 1, cno))
				if(sendword(XYM_steps, cno)){
					sleep(1);
					ret = 1;
//					send_cmd(SET12V, 1, cno);
				}
		}
	return ret;
}

bool goto_middle(){
	bool ret = 1;
	WARN("   \n");
	WARN("  X... ");
	if(!gtmid(X_Motor)){
		WARN("!\n");
		ret = 0;
	}
	else{
		WARN("OK\n");
		X_pos = 0;
	}
	WARN("  Y... ");
	if(!gtmid(Y_Motor)){
		WARN("!\n");
		ret = 0;
	}
	else{
		WARN("OK\n");
		Y_pos = 0;
	}
	if(ret){Xprev = 0; Yprev = 0;}
	return ret;
}

void move_to_center(){ 
	Guiding_Lock = 1;
//	init_XY();
	if(!goto_middle()){
		WARN("!   ...\n");
		goto_middle();
	}
//	sleep(1);
	Guiding_Lock = 0;
}

bool init_motor(unsigned char cno){
//	unsigned int mask = 37262;// 0b1001000110001110; //   ,
	/*         :
	  |TS|  A | stop_mask ||TF|  B | mode_mask |
	H/| 7| 6 5| 4 3 2 1 0 ||7 | 6 5| 4 3 2 1 0 |/L
	TS - terminal_stop_motor (1-  )
	TF - terminal_flag (1- )
	mode_mask -     
	stop_mask - 1-     1
	AB-stop_mask2 - 1-   2 
	*/
	ERR(" (%s)\n", cnameByNum(cno));
	if(!init_ctrlr(cno)) return 0;
	if(!send_cmd(ISSTEP, 1, cno)) return 0;
	if(!send_cmd(SET24V, 1, cno)) return 0;
	if(send_cmd(SET_TIMER, 1, cno)){
		if(!sendword(65376, cno)){ // 12.8
			WARN("  (%d)\n", cnameByNum(cno));
			return 0;
		}
	} else return 0;
/*	if(cno == C_Motor){ //   -    
		if(send_cmd(TERM_METHOD, 1, cno)){
			if(!sendword(159, cno)){
				WARN("    \n");
				return 0;
			}
		} else return 0;
		Shutter_lock = 0;
	}*/
//	send_cmd(TERM_METHOD, 1, cno);
//	if(!sendword(mask, cno)) WARN("    %d\n", cno);
	return 1;
}
	
void init_XY(){
	char data[256];
	bool res = 0, ini = 1;
	tty_init();
	inline void test(bool res){
		if(!res){
			WARN(" \n");
			ini = 0;
		} else WARN("\n");	
	}
/*	WARN("  : ");
	//    
	if(res = init_ctrlr(L_Motor))
	if(res = send_cmd(ISSTEP, 1, L_Motor))
		if(res = send_cmd(SET24V, 1, L_Motor))
			if(res = send_cmd(SET_TIMER, 1, L_Motor))
				if(sendword(65286, L_Motor) == 0){ 
					WARN("  \n"); // 2
					res = 0;
				}
	test(res);
	//   
	if(res = init_ctrlr(Lamps))
	if(res = send_cmd(ISNTSTEP, 1, Lamps))
		if(res = send_cmd(SET12V, 1, Lamps)){
			manage_lamps(0); //  
			manage_lamps(8); //     
			WARN(" : \n");
		}
	test(res);
	*/
	//   
	if(res = init_ctrlr(Shutt))
	if(res = send_cmd(ISNTSTEP, 1, Shutt)){
		shutter(1);
		WARN(" : \n");
	}
	test(res);
	//  XY-
/*	if(!init_motor(Y_Motor)){ //  Y
		WARN("Y:  !\n");
		ini = 0;
	}
	if(!init_motor(X_Motor)){ //  X
		WARN("X:  !\n");
		ini = 0;
	}
*/	inline void setzero(unsigned int cno, int var){
		if(var < 1){
			if(send_cmd(STEP_L_NUM, 1, cno))
				sendword(630, cno);
		}
	}
	//    
/*	if(!init_motor(G_Motor)){
		WARN(" :  !");
		ini = 0;
	}else setzero(G_Motor, newGrPos);*/
	//     
	if(!init_motor(C_Motor)){
		WARN("  :  !");
		ini = 0;		
	}else setzero(C_Motor, CamPos);
	//   
/*	if(!init_motor(Focus)){
		WARN(" :  !");
		ini = 0;		
	}else setzero(Focus, newPh);*/
	if(ini) WARN("  \n");
	else WARN("   \n");
}

void step_goto(double *xold, double *yold, double xnew, double ynew){
	double StepsX, StepsY;
	int Sx, Sy;
	double dr_x, dr_y;
	unsigned char command;
	if(*xold == CH_OBJ || *yold == CH_OBJ){ //   
		*xold = 0;
		*yold = 0;
		return;
	}
	//dr_x = -xnew + *xold;
	dr_x =  *xold - xnew;
	StepsX = dr_x / ScaleX * FeedBack; // FeedBack -   
	dr_y = *yold - ynew;
	StepsY = dr_y / ScaleY * FeedBack;
	//     
	Sx = (int)(StepsX * COS_TH - StepsY * SIN_TH);
	Sy = (int)(StepsX * SIN_TH + StepsY * COS_TH);
	
	if(abs(Sx) > MIN_STEPS){ //   X -  0
		if(Sx > 0) command = 1; 
		else command = 0;
		if((abs(X_pos + Sx) < XYM_steps)){
			MSG("\n    %d \n", Sx);
			if(corrector(abs(Sx), command)){
				XSteps = Sx; //    
				X_pos += Sx; //   ( )
			}
			else Sx = 0;
		}
		else{
			WARN("X:   \n");
			Sx = 0;
		}
	}
	else Sx = 0;

	if(abs(Sy) > MIN_STEPS){ //   Y -  1
		if(Sy > 0) command = 3;
		else command = 2;
		if((abs(Y_pos + Sy) < XYM_steps)){
			MSG("\n  Y  %d \n", Sy);
			if(corrector(abs(Sy), command)){
				YSteps = Sy; //  
				Y_pos += Sy;
			}
			else Sy = 0;
		}
		else{
			WARN("Y:   \n");
			Sy = 0;
		}
	}
	else Sy = 0;

	*xold += ScaleX*(Sx * COS_TH + Sy * SIN_TH);
	*yold += ScaleY*(Sy * COS_TH - Sx * SIN_TH);
	MSG("move: xold=%.1f, yold=%.1f; xnew=%.1f, ynew=%.1f\n", *xold, *yold, xnew, ynew);
}

void stop_all_devices(){
	unsigned char i;
	WARN("  \n");
	if(comfd > 0){
		for(i=0; i<8; i++)
			send_cmd(STEP_STOP, 1, i<<5); //  
		ioctl(comfd, TCSETA, &oldtty); //    com
		close(comfd); //  
	}
}

void manage_lamps(char lamp){
/*   lamp ( -  ,  -  ):
 * 0 -   (DPM12_STOP )
 * 1 -   1 (DPM1R )
 * 2 -   2 (DPM2R )
 * 4 -   (STEP_L_NUM  LM_Steps  )
 * 5 -     1 (1+4)
 * 6 -     2 (2+4)
 * 8 -  (STEP_R_NUM  LM_Steps  )
 *   
 */
 	unsigned char lamp_cmd = 0, motor_cmd = 0;
	switch(lamp){
		case 0: lamp_cmd = DPM12_STOP; break;
		case 5: motor_cmd = STEP_L_NUM; //  break  !
		case 1: lamp_cmd = DPM1R; break;
		case 6: motor_cmd = STEP_L_NUM; //  break  !
		case 2: lamp_cmd = DPM2R; break;
		case 4: motor_cmd = STEP_L_NUM; break;
		case 8: motor_cmd = STEP_R_NUM; break;
	}
	if(lamp_cmd){
		send_cmd(DPM12_STOP, 1, Lamps); //    
		if(lamp)send_cmd(lamp_cmd, 1, Lamps);
	}
	if(motor_cmd)
		if(send_cmd(motor_cmd, 1, L_Motor))
			sendword(LM_Steps, L_Motor);	
}

void shutter(bool flag){
	usleep(10000);
	send_cmd(SET12V, 1, Shutt);
	usleep(10000);
	if(flag){// 
//		if(send_cmd(DPM1L2L, 1, Shutt)) oldshtr = 1;
		if(send_cmd(DPM12_STOP, 1, Shutt)) oldshtr = 1;
	}
	else{
		if(send_cmd(DPM1R2R, 1, Shutt)) oldshtr = 0;
	}
	usleep(10000);
	send_cmd(SET09V, 1, Shutt);
}

void change_pos(int newpos, int *curpos, unsigned char cno){
	int dpos, sign = 1;
	unsigned char cmd;
	if(newpos < 0) newpos = 0;
	if(newpos > 630) newpos = 630;
	dpos = newpos - *curpos;
	if(dpos < 0){dpos *= -1; sign = -1;}
	cmd = (sign > 0) ? STEP_R_NUM : STEP_L_NUM;
	if(dpos)
		if(send_cmd(cmd, 1, cno))
			if(sendword(dpos, cno))
				*curpos = newpos;
}

void *controllers_thread(){ // ,  
	bool guiding, movetocenter, init, shtr;
	char lamp, oldlamp=0, dcam, dgr, dph;
	double xc, yc, tnow;
	int r, newcampos, dx, dy, newgmotor, newphocus;
	if((r = pthread_mutex_lock(&mutex)) != 0)
		WARN("mutex_lock motor_f (%d), %s\n", strerror(r));
	Shutter_lock = 0;
	Shutter = 0;
	Init = 0;
	Lamp = 0;
	MOVETOCENTER = 0;
	Guiding = 0;
	Guiding_Lock = 0;
	if((r = pthread_mutex_unlock(&mutex)) != 0)
		WARN("mutex_unlock motor_f (%d), %s\n", strerror(r));
	init_XY();
	move_to_center();
	WARN(" \n");
	for(;;){
		if(read_ctrlr()) continue; //   
		if((r = pthread_mutex_trylock(&mutex)) != 0){
			WARN("mutex_trylock (%d): %s\n", r, strerror(r));
			continue; // ,    
		}
		guiding = Guiding;
		xc = xcent; yc = ycent;
		movetocenter = MOVETOCENTER;
		lamp = Lamp;
		init = Init;
		shtr = Shutter;
		newcampos = newCamPos; dcam = dCam;
		dx = dX; dy = dY;
		newgmotor = newGrPos; newphocus = newPh;
		dcam = dCam; dgr = dGr; dph = dPh;
		if(Shutter_lock){
			tnow = dtime() - tbeg;
			if((int)tnow > ExpTime + 1)
				ExpTime = (int) tnow;
		}
		//     
		if((r = pthread_mutex_unlock(&mutex)) != 0)
			WARN("mutex_unlock (%d), %s\n", strerror(r));
		if(init || movetocenter){
			pthread_mutex_lock(&mutex);
			Guiding = 0;
			Guiding_Lock = 1;
			MOVETOCENTER = 0;
			Init = 0;
			pthread_mutex_unlock(&mutex);
		}
		if(init){
			init_XY();
			continue;
		}
		if(movetocenter){
			move_to_center();
			continue;
		}
		if(dx || dy){
			dX=0; dY=0;
			double xx=0, yy=0;
			//if(!Guiding_Lock) step_goto(&Xprev, &Yprev, Xprev+dx/0.9, Yprev+dy/0.9);
			if(!Guiding_Lock){
				if(!guiding) step_goto(&xx, &yy, dx/FeedBack, dy/FeedBack);
				else {Xprev -= dx; Yprev -= dy;}
			}
		}
		if(guiding && !Guiding_Lock)
			step_goto(&Xprev, &Yprev, xc, yc); //   - X,Yprev    
		if(lamp != oldlamp){
			oldlamp = lamp;
			manage_lamps(lamp);
		}
		if(shtr != oldshtr && !Shutter_lock)
			shutter(shtr);
		if(newcampos>-1){change_pos(newcampos, &CamPos, C_Motor); newCamPos = -1;}
		if(dcam){change_pos(CamPos+dcam, &CamPos, C_Motor); dCam = 0;}
		if(newgmotor>-1){change_pos(newgmotor, &GrPos, G_Motor); newGrPos = -1;}
		if(dgr){change_pos(GrPos+dgr, &GrPos, G_Motor); dGr = 0;}
		if(newphocus>-1){change_pos(newphocus, &Ph, Focus); newPh = -1;}
		if(dph){change_pos(Ph+dph, &Ph, Focus); dPh = 0;}
	}
}

