//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define V4L2CONTROL_C
#include "v4l2control.h"

struct buffer * buffers = NULL;
unsigned int n_buffers = 0;
int  grab_fd=-1;
char* grab_dev = GRAB_DEVICE;
double* filterbuf;
unsigned char *images[4];

typedef enum {
	IO_METHOD_READ,
	IO_METHOD_MMAP,
	IO_METHOD_USERPTR,
} io_method;
static io_method io = IO_METHOD_MMAP;

void errno_exit (const char * s){
	WARN("%s: error %d, %s\n", s, errno, strerror(errno));
	printf("%s: error %d, %s\n", s, errno, strerror(errno));
	rm_queues();
	exit(EXIT_FAILURE);
}

int xioctl (int grab_fd, int request, void * arg){
	int r;
	do r = ioctl (grab_fd, request, arg);
	while (-1 == r && EINTR == errno);
	return r;
}

/*
		   
*/

static bool init_read(unsigned int buffer_size){
	buffers = calloc(1, sizeof (*buffers));
	if (!buffers){
		ERR("init_read: Out of memory");
		return 0;
	}
	buffers[0].length = buffer_size;
	buffers[0].start = malloc(buffer_size);
	if (!buffers[0].start){
		ERR("init_read: Out of memory");
		return 0;
	}
	WARN("\ninit_read, buf size = %d\n", buffer_size);
	return 1;
}

static bool init_mmap(unsigned int *bufsize){
	struct v4l2_requestbuffers req;
	CLEAR (req);
	req.count = 4;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_MMAP;
	if(-1 == xioctl (grab_fd, VIDIOC_REQBUFS, &req)){
		if(EINVAL == errno){
			ERR("%s does not support memory mapping\n", grab_dev);
			return 0;
		}else{
			ERR("VIDIOC_REQBUFS");
			return 0;
		}
	}
	if(req.count < 2){
		ERR("Insufficient buffer memory on %s\n", grab_dev);
		return 0;
	}
	buffers = calloc(req.count, sizeof (*buffers));
	if(!buffers){
		ERR("Out of memory");
		return 0;
	}
	for(n_buffers = 0; n_buffers < req.count; ++n_buffers){
		struct v4l2_buffer buf;
		CLEAR (buf);
		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;
		buf.index = n_buffers;
		if(-1 == xioctl(grab_fd, VIDIOC_QUERYBUF, &buf)){
			ERR("VIDIOC_QUERYBUF");
			return 0;
		}
		buffers[n_buffers].length = buf.length;
		buffers[n_buffers].start = mmap(NULL, // start anywhere
			buf.length,
			PROT_READ | PROT_WRITE, // required
			MAP_SHARED, // recommended
			grab_fd, buf.m.offset);
		if(MAP_FAILED == buffers[n_buffers].start){
			ERR("mmap");
			return 0;
		}
	}
	WARN("\ninit_mmap(), nbuffers=%d, buf size=%d\n", sizeof (*buffers), buffers[0].length);
	return 1;
}

static bool init_userp(unsigned int buffer_size){
	struct v4l2_requestbuffers req;
	unsigned int page_size;
	page_size = getpagesize();
	buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);
	CLEAR (req);
	req.count = 4;
	req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	req.memory = V4L2_MEMORY_USERPTR;
	if(-1 == xioctl(grab_fd, VIDIOC_REQBUFS, &req)){
		if(EINVAL == errno){
			ERR("%s does not support user pointer i/o\n", grab_dev);
			return 0;
		}else{
			ERR("VIDIOC_REQBUFS");
			return 0;
		}
	}
	buffers = calloc(4, sizeof (*buffers));
	if(!buffers){
		ERR("Out of memory");
		return 0;
	}
	for(n_buffers = 0; n_buffers < 4; ++n_buffers){
		buffers[n_buffers].length = buffer_size;
		buffers[n_buffers].start = images[n_buffers];
		if(!buffers[n_buffers].start){
			ERR("Out of memory");
			return 0;
		}
	}
	WARN("\ninit_userp, buf size=%d, n buffers = %d\n", buffer_size, n_buffers);
	return 1;
}

static void init_buffers(unsigned int buffer_size){
	key_t tkey;
	int shm_new = 0, i;
	char shmname[32];
	int filbufsize = 65536 * sizeof(double); //    . .
	int fd;
	unsigned int page_size = getpagesize();
	buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);
	//     
	tkey = ftok(grab_dev, 666);
	shmid = shmget(tkey, sizeof(struct Buffs), IPC_CREAT| IPC_EXCL | 0666);
	if(shmid == -1){
		if( (shmid = shmget(tkey, sizeof(struct Buffs), 0666)) == -1)
			errno_exit("   ");
	}
	else
		shm_new = 1;
	Buffers = (struct Buffs*) shmat(shmid, NULL, 0);
	if(Buffers == (void *) -1) errno_exit("   ");
	if(shm_new) Buffers->nbuf_cur = -1;
	for(i = 0; i < 4; i++){
		snprintf(shmname, 31, "/shm%d", i);
		fd = shm_open(shmname, O_RDWR|O_CREAT, 00666);
		if(fd < 0){shm_unlink(shmname); errno_exit("      ");}
		ftruncate(fd, buffer_size);
		images[i] = (unsigned char*)mmap(NULL, buffer_size, PROT_READ|PROT_WRITE, 
			MAP_SHARED, fd, 0);
		if (images[i] == MAP_FAILED)
			errno_exit("    ");
	}
	fd = shm_open("/shm5", O_RDWR|O_CREAT, 00666);
	if(fd < 0){shm_unlink(shmname); errno_exit("      ");}
	ftruncate(fd, filbufsize);
	filterbuf = mmap(NULL, filbufsize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (filterbuf == MAP_FAILED) errno_exit("     ");
}

static void start_capturing (void){
	unsigned int i;
	enum v4l2_buf_type type;
	switch(io){
	case IO_METHOD_READ:
		break;

	case IO_METHOD_MMAP:
		for(i = 0; i < n_buffers; ++i){
			struct v4l2_buffer buf;
			CLEAR (buf);
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_MMAP;
			buf.index = i;
			if(-1 == xioctl(grab_fd, VIDIOC_QBUF, &buf))
				errno_exit("VIDIOC_QBUF");
		}
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if(-1 == xioctl(grab_fd, VIDIOC_STREAMON, &type))
			errno_exit("VIDIOC_STREAMON");
		break;

	case IO_METHOD_USERPTR:
		for(i = 0; i < n_buffers; ++i){
			struct v4l2_buffer buf;
			CLEAR (buf);
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_USERPTR;
			buf.index = i;
			buf.m.userptr = (unsigned long) buffers[i].start;
			buf.length = buffers[i].length;
			if(-1 == xioctl(grab_fd, VIDIOC_QBUF, &buf))
				errno_exit("VIDIOC_QBUF");
		}
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if(-1 == xioctl(grab_fd, VIDIOC_STREAMON, &type))
			errno_exit("VIDIOC_STREAMON");
		break;
	}
	WARN("\nstart_capturing()\n");
}

void grab_init(char *dev){
	struct v4l2_capability cap;
	struct v4l2_cropcap cropcap;
	struct v4l2_crop crop;
	struct v4l2_format fmt;
	int i = 0;
	bool ret = 0;
	unsigned int bufsize;
	if(grab_fd < 0){ //    
		struct stat st; 
		if(-1 == stat(dev, &st)){
			ERR("Cannot identify '%s': %d, %s\n", grab_dev, errno, strerror (errno));
			exit (EXIT_FAILURE);
		}
		if(!S_ISCHR(st.st_mode)){
			ERR("%s is no device\n", grab_dev);
			exit(EXIT_FAILURE);
		}
		grab_fd = open(dev, O_RDWR | O_NONBLOCK, 0);
		if(-1 == grab_fd){
			ERR("Cannot open '%s': %d, %s\n", grab_dev, errno, strerror (errno));
			exit(EXIT_FAILURE);
		}
	}
	if(-1 == xioctl(grab_fd, VIDIOC_QUERYCAP, &cap)){
		if(EINVAL == errno){
			ERR("%s is not V4L2 device\n", grab_dev);
			exit(EXIT_FAILURE);
		} else
			errno_exit ("VIDIOC_QUERYCAP");
	}
	if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)){
		ERR("%s is no video capture device\n", grab_dev);
		exit (EXIT_FAILURE);
	}
	for(i = 0; i<2; i++){
		switch(io){
		case IO_METHOD_READ:
			if (!(cap.capabilities & V4L2_CAP_READWRITE)){
				ERR("%s does not support read i/o, trying streaming\n", grab_dev);
				io = IO_METHOD_MMAP;
			}
			else			
				i = 2;
			break;

		case IO_METHOD_MMAP:
		case IO_METHOD_USERPTR:
			if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
				ERR("%s does not support streaming i/o, trying reading\n", grab_dev);
				io = IO_METHOD_READ;
			}
			else
				i = 2;
			break;
		}
	}
	// Select video input, video standard and tune here.
	grab_set_chan(GrabChan);
	CLEAR (cropcap);
	cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if(0 == xioctl (grab_fd, VIDIOC_CROPCAP, &cropcap)){
		crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		crop.c = cropcap.defrect; // reset to default 
		if(-1 == xioctl(grab_fd, VIDIOC_S_CROP, &crop)){
			switch (errno){
			case EINVAL:
				WARN("Cropping not supported\n");
			break;
			default:
				WARN("VIDIOC_S_CROP: error\n");
			break;
			}
		}
	} else { 
		WARN("VIDIOC_CROPCAP error\n");
	}
	CLEAR (fmt);
	fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	fmt.fmt.pix.width = Grab_Width; 
	fmt.fmt.pix.height = Grab_Height;
	fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_GREY; 
	fmt.fmt.pix.field = V4L2_FIELD_INTERLACED; 
	if(-1 == xioctl(grab_fd, VIDIOC_S_FMT, &fmt))
		errno_exit("VIDIOC_S_FMT");
	Grab_Width = fmt.fmt.pix.width;
	Grab_Height = fmt.fmt.pix.height;
	bufsize = fmt.fmt.pix.sizeimage;
	init_buffers(bufsize);
	for(i=0; i<3; i++){ //   ,    
		switch(io){
		case IO_METHOD_READ:
			if(ret = init_read(bufsize)) i = 3;
			else io = IO_METHOD_USERPTR;
			break;
		case IO_METHOD_MMAP:
			if(ret = init_mmap(&bufsize)) i = 3;
			else io = IO_METHOD_READ;
			break;
		case IO_METHOD_USERPTR:
			if(ret = init_userp(bufsize)) i = 3;
			else io = IO_METHOD_MMAP;
			break;
		}
	}
	Buffers->length = buffers[0].length;
	if(!ret){
		errno_exit("No methods available");
	}
	WARN("grab_init(): format=%dx%d\n", Grab_Width, Grab_Height);
	start_capturing();
}

//   
void grab_set_chan(int ch_num){
	if (-1 == ioctl (grab_fd, VIDIOC_S_INPUT, &ch_num)){
		errno_exit("VIDIOC_S_INPUT");
	}
	MSG("\ngrab_set_chan(%d)\n", ch_num);
}

void change_controls(int brightness, int contrast){
	struct v4l2_queryctrl queryctrl;
	struct v4l2_control control;
	if (brightness != 0){
		memset (&queryctrl, 0, sizeof (queryctrl));
		queryctrl.id = V4L2_CID_BRIGHTNESS;
		if (-1 == ioctl (grab_fd, VIDIOC_QUERYCTRL, &queryctrl)) {
			if (errno != EINVAL) {
				perror ("VIDIOC_QUERYCTRL");
				exit (EXIT_FAILURE);
			} else {
				printf ("V4L2_CID_BRIGHTNESS is not supported\n");
			}
		} else if (queryctrl.flags & V4L2_CTRL_FLAG_DISABLED) {
			printf ("V4L2_CID_BRIGHTNESS is not supported\n");
		} else {
			memset (&control, 0, sizeof (control));
			control.id = V4L2_CID_BRIGHTNESS;
			control.value = brightness; //queryctrl.default_value;
			if (-1 == ioctl (grab_fd, VIDIOC_S_CTRL, &control)) {
				perror ("VIDIOC_S_CTRL");
				exit (EXIT_FAILURE);
			}
		}
	}
	if (contrast != 0){
		memset (&control, 0, sizeof (control));
		control.id = V4L2_CID_CONTRAST;
		if (0 == ioctl (grab_fd, VIDIOC_G_CTRL, &control)) {
			control.value = contrast;//+= 1;
		/* The driver may clamp the value or return ERANGE, ignored here */
			if (-1 == ioctl (grab_fd, VIDIOC_S_CTRL, &control) && errno != ERANGE) {
				perror ("VIDIOC_S_CTRL");
				exit (EXIT_FAILURE);
			}
		/* Ignore if V4L2_CID_CONTRAST is unsupported */
		} else if (errno != EINVAL) {
			perror ("VIDIOC_G_CTRL");
			exit (EXIT_FAILURE);
		}
	}
}

int fillBuf(unsigned char* img){ //    ( MMAP, READ)
	static int CurBuf = -1;
	int i, sz = Grab_Width * Grab_Height;
	unsigned char *ptr;
	if(++CurBuf > 4) CurBuf = 0;
	ptr = images[CurBuf];
	for(i = 0; i < sz; i++)
		*ptr++ = *img++;
	return CurBuf;
}

//    (~100 100100)
unsigned char* grab_next(int *width, int *height){
	struct v4l2_buffer buf;
	unsigned int i;
	void* ret;
	fd_set fds;
	struct timeval tv;
	int r;
#ifdef __MORE_TEST__
	double t0 = dtime();
#endif
	for(;;){
		FD_ZERO(&fds);
		FD_SET(grab_fd, &fds);
		tv.tv_sec = 2;
		tv.tv_usec = 0;
		r = select (grab_fd + 1, &fds, NULL, NULL, &tv);
		if (-1 == r){
			if (EINTR == errno)
			continue;
			errno_exit ("select");
		}
		if (0 == r){
			errno_exit("select timeout");
		}
		switch(io){

		case IO_METHOD_READ:
			if(-1 == read(grab_fd, buffers[0].start, buffers[0].length)){
				switch(errno){
					case EAGAIN:
						return NULL;
					case EIO:
						ERR("EIO");
						break;
					default:
					errno_exit("read");
				}
			}
			ret = buffers[0].start;
			Buffers->nbuf_cur = fillBuf(ret);
			break;

		case IO_METHOD_MMAP:
			CLEAR (buf);
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_MMAP;
			if(-1 == xioctl(grab_fd, VIDIOC_DQBUF, &buf)){
				switch(errno){
					case EAGAIN:
						return NULL;
					case EIO:
						ERR("EIO");
						break;
					default:
						errno_exit("VIDIOC_DQBUF");
				}
			}
			assert(buf.index < n_buffers);
			ret = buffers[buf.index].start;
			if(-1 == xioctl(grab_fd, VIDIOC_QBUF, &buf))
				errno_exit("VIDIOC_QBUF");
			Buffers->nbuf_cur = fillBuf(ret);
			break;

		case IO_METHOD_USERPTR:
			CLEAR (buf);
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_USERPTR;
			if(-1 == xioctl(grab_fd, VIDIOC_DQBUF, &buf)){
				switch (errno){
					case EAGAIN:
						return NULL;
					case EIO:
						ERR("EIO");
						break;
					default:
					errno_exit("VIDIOC_DQBUF");
				}
			}
			for(i = 0; i < n_buffers; ++i)
				if(buf.m.userptr == (unsigned long) buffers[i].start
					&& buf.length == buffers[i].length)
					break;
			assert(i < n_buffers);
			Buffers->nbuf_cur = buf.index;
			ret = (void *) buf.m.userptr;
			if(-1 == xioctl(grab_fd, VIDIOC_QBUF, &buf))
				errno_exit("VIDIOC_QBUF");
			break;
		}
		if (ret != NULL) break;
	}
	MSG("\ngrab_next(): capture time=%f, ", dtime()-t0);
	*width = Grab_Width;
	*height = Grab_Height;
	return (unsigned char*) ret;
}

