//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.

#define __WAVECORR_C__
#include "iftest.h"
#include "fiber_guide_cgi.h" //  dtime, time0
#include "wavecorr.h"

void wavefilt(double *data, int size){
	int i, x, y;
	gsl_wavelet2d_transform_forward(w, data, 1, size, size, work); //  stride = 0
	i = size*size;
	for(x=0; x<i; x++)
		abscoeff[x] = fabs(data[x]);
	gsl_sort_index(p, abscoeff, 1, i);
	y = i - W_filter.treshold;
	for(x=0; x<y; x++)
		data[p[x]] = 0;
	gsl_wavelet2d_transform_inverse(w, data, 1, size, size, work); //  stride = 0
}

void Realloc_memory(int size){
	int s2 = size * size;
	MSG("Realloc memory..");
	if(CCF != NULL){ //  ,    
		gsl_wavelet_free(w),
		gsl_wavelet_workspace_free(work);
		free(abscoeff);
		free(p);
		fftw_destroy_plan(im0);
		fftw_destroy_plan(im1);
		fftw_destroy_plan(rev);
		fftw_free(fft0); fftw_free(fft1); fftw_free(fft2);
		free(CCF);
	}
	abscoeff = (double*) malloc(s2 * sizeof(double));
	p = (size_t*) malloc(s2 * sizeof(size_t));
	CCF = (double*) malloc(s2 * sizeof(double));

	fft0 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * s2);
	fft1 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * s2);
	fft2 = (fftw_complex*)fftw_malloc(sizeof(fftw_complex) * s2);

	im0 = fftw_plan_dft_r2c_2d(size, size, img0, fft0, FFTW_PATIENT);
	im1 = fftw_plan_dft_r2c_2d(size, size, img, fft1, FFTW_PATIENT);
	rev = fftw_plan_dft_c2r_2d(size, size, fft2, CCF, FFTW_PATIENT);
	//   
	w = gsl_wavelet_alloc(gsl_wavelet_daubechies, W_filter.order); //   order=4
	work = gsl_wavelet_workspace_alloc(size); 
	MSG(" size=%d, done.\n", size);
}

void filter_image(filter_type type, double* image, int size){
	int i, ss = size*size;
	double *fptr = filterbuf, *iptr = image;
	switch(type){
		case Wavelet:
			wavefilt(image, size);
		break;
		case Median:
			medfilt(image, size);
		break;
	}
	if(F_control && (size < 257)) //      
		for(i = 0; i < ss; i++)
			*fptr++ = *iptr++;
}

void makeRefImage(filter_type type, int size){
	int i, j;
	double *bp, *bp0;
	bp = img; bp0 = img0;
	for(j=0; j<size; j++)
		for(i=0; i<size; i++)
		    *bp0++ = *bp++;
	filter_image(type, img0, size);
	fftw_execute(im0);
}

void findcenter(filter_type type, double *xc, double *yc, int size){
	int x, y, i, xx = 0, yy = 0;
	double xcn=0., ycn=0., xpar, ypar;
	double a, b, c, d, *w, I = 0., Imax = 0., tmp;
	filter_image(type, img, size);
	if(CMass){ // ""  ,  ~50%  S/N=50; ~35% S/N=100; ~10% S/N=500
		double max, min, med, tres;
		i = size * size;
		med = quick_select(img, i); //  
		max = min = *img;
		w = img;
		for(x = 1; x < i; x++){ //  
			tmp = *(++w);
			if(max < tmp) max = tmp;
			else if(min > tmp) min = tmp;
		}
		//     med - 0.1(max-min),    0.1(max-min)
		tmp = 0.1 * (max - min);
		tres = med - tmp;
		if(tres < tmp) tres = tmp;
		w = img;
		for(y = 0; y < size; y++)
			for(x = 0; x < size; x++)
				if((d = (*(w++) - tres)) > 0.){
					xcn += d * x;
					ycn += d * y;
					I += d;
				}
	}
	else{ // ,   S/N=10: ~50%  ,
			// ~30%   , ~10%  -
		fftw_execute(im1);
		//  (cplx)fft0 *= fft1
		for(i=x=0; x<size; x++)
		for(y=0; y<size; y++, i++){
			a=fft0[i][0]; c=fft1[i][0];
			b=fft0[i][1]; d=fft1[i][1];
			fft2[i][0] = a*c + b*d;
			fft2[i][1] = b*c - a*d;
			tmp = 1.;
			if((y&1) == 1) tmp = -tmp; //  
			fft2[i][0] /= tmp;
			fft2[i][1] /= tmp;
		}	//   ( )
		fftw_execute(rev);
		w = CCF;
		for(y=0; y<size; y++){
			for(x=0; x<size; x++){
				if((d = *w++) > Imax){
					Imax = d;
					xx = x;
					yy = y;
				}
			}
		}
		MSG("findcenter: xx=%d, yy=%d, CCF[yy,xx]=%.1f\n", xx, yy, CCF[yy*size+xx]);
		for(y=yy-3; y<yy+4; y++)
			for(x=xx-3; x<xx+4; x++){
				d = CCF[y*size+x];
				if(d < 0) continue;
				xcn += x * d;
				ycn += y * d;
				I += d;
			}
	}
	d = size/2.;
	if(I>1.){	
		xcn /= I;
		ycn /= I;
	}
	else {xcn = d; ycn = d;}
	*xc = xcn - d;
	//*xc = 0.;
	*yc = ycn - d;
	if(!CMass){ *yc *= -1.0;}// *xc *= -1.;}
	else *xc *= -1.0;
	MSG("xcn=%.1f, ycn=%.1f, I=%.1f, d=%.2f, xc=%.1f, yc=%.1f\n", xcn, ycn, I, d,*xc,*yc);
}

